#ifndef __NRF24L01
#define __NRF24L01

// library for the NRF24L01 transceiver
// ported from the RF24 arduino library at https://github.com/maniacbug/RF24

#include <stdint.h>
#include <plib.h>

#define dTime_ms 60000000/5000
#define dTime_us 60000000/5000000

/* Memory Map */
#define RF24_CONFIG      0x00
#define RF24_EN_AA       0x01
#define RF24_EN_RXADDR   0x02
#define RF24_SETUP_AW    0x03
#define RF24_SETUP_RETR  0x04
#define RF24_RF_CH       0x05
#define RF24_RF_SETUP    0x06
#define RF24_NRF_STATUS  0x07
#define RF24_OBSERVE_TX  0x08
#define RF24_CD          0x09
#define RF24_RX_ADDR_P0  0x0A
#define RF24_RX_ADDR_P1  0x0B
#define RF24_RX_ADDR_P2  0x0C
#define RF24_RX_ADDR_P3  0x0D
#define RF24_RX_ADDR_P4  0x0E
#define RF24_RX_ADDR_P5  0x0F
#define RF24_TX_ADDR     0x10
#define RF24_RX_PW_P0    0x11
#define RF24_RX_PW_P1    0x12
#define RF24_RX_PW_P2    0x13
#define RF24_RX_PW_P3    0x14
#define RF24_RX_PW_P4    0x15
#define RF24_RX_PW_P5    0x16
#define RF24_FIFO_STATUS 0x17
#define RF24_DYNPD	     0x1C
#define RF24_FEATURE	 0x1D

/* Bit Mnemonics */
#define RF24_MASK_RX_DR  6
#define RF24_MASK_TX_DS  5
#define RF24_MASK_MAX_RT 4
#define RF24_EN_CRC      3
#define RF24_CRCO        2
#define RF24_PWR_UP      1
#define RF24_PRIM_RX     0
#define RF24_ENAA_P5     5
#define RF24_ENAA_P4     4
#define RF24_ENAA_P3     3
#define RF24_ENAA_P2     2
#define RF24_ENAA_P1     1
#define RF24_ENAA_P0     0
#define RF24_ERX_P5      5
#define RF24_ERX_P4      4
#define RF24_ERX_P3      3
#define RF24_ERX_P2      2
#define RF24_ERX_P1      1
#define RF24_ERX_P0      0
#define RF24_AW          0
#define RF24_ARD         4
#define RF24_ARC         0
#define RF24_PLL_LOCK    4
#define RF24_RF_DR       3
#define RF24_RF_PWR      6
#define RF24_RX_DR       6
#define RF24_TX_DS       5
#define RF24_MAX_RT      4
#define RF24_RX_P_NO     1
#define RF24_TX_FULL     0
#define RF24_PLOS_CNT    4
#define RF24_ARC_CNT     0
#define RF24_TX_REUSE    6
#define RF24_FIFO_FULL   5
#define RF24_TX_EMPTY    4
#define RF24_RX_FULL     1
#define RF24_RX_EMPTY    0
#define RF24_DPL_P5	     5
#define RF24_DPL_P4	     4
#define RF24_DPL_P3	     3
#define RF24_DPL_P2	     2
#define RF24_DPL_P1	     1
#define RF24_DPL_P0	     0
#define RF24_EN_DPL	     2
#define RF24_EN_ACK_PAY  1
#define RF24_EN_DYN_ACK  0

/* Instruction Mnemonics */
#define RF24_R_REGISTER    0x00
#define RF24_W_REGISTER    0x20
#define RF24_REGISTER_MASK 0x1F
#define RF24_ACTIVATE      0x50
#define RF24_R_RX_PL_WID   0x60
#define RF24_R_RX_PAYLOAD  0x61
#define RF24_W_TX_PAYLOAD  0xA0
#define RF24_W_ACK_PAYLOAD 0xA8
#define RF24_FLUSH_TX      0xE1
#define RF24_FLUSH_RX      0xE2
#define RF24_REUSE_TX_PL   0xE3
#define RF24_NOP           0xFF

#define RF24_LNA_HCURR     0

#define RF24_RPD           0x09
#define RF24_W_TX_PAYLOAD_NO_ACK 0xB0

#define RF24_RF_DR_LOW   5
#define RF24_RF_DR_HIGH  3
#define RF24_RF_PWR_LOW  1
#define RF24_RF_PWR_HIGH 2


// macros to control CE and CSN
#define _rf24_ce_high() {LATBSET = (1 << 2);}
#define _rf24_ce_low()  {LATBCLR = (1 << 2);}

#define _rf24_cs_high() {LATBSET = (1 << 1);}
#define _rf24_cs_low()  {LATBCLR = (1 << 1);}

#define rf24_min(a,b) (a<b?a:b)
#define rf24_max(a,b) (a>b?a:b)

#define _BV(x) (1<<(x))

// power amplifier level
// set using rf24_set_amplifier_level()
typedef enum { RF24_PA_MIN = 0,RF24_PA_LOW, RF24_PA_HIGH, RF24_PA_MAX, RF24_PA_ERROR } rf24_pa_dbm_e;

// data rate
// set using rf24_set_data_rate()
typedef enum { RF24_1MBPS = 0, RF24_2MBPS, RF24_250KBPS } rf24_datarate_e;

// crc length
// set using rf24_set_crc_length()
typedef enum { RF24_CRC_DISABLED = 0, RF24_CRC_8, RF24_CRC_16 } rf24_crclength_e;

uint8_t rf24_wide_band; // indicates if 2mbps data rate is in use
uint8_t rf24_p_variant; // 0 for RF24L01 and 1 for RF24L01P
uint8_t rf24_payload_size; // fixed size of payloads
uint8_t rf24_ack_payload_available; // whether there is an ack payload waiting
uint8_t rf24_dynamic_payloads_enabled; // flag for dynamic payload sizes enabled.
uint8_t rf24_ack_payload_length; // dynamic size of pending ack payload
uint8_t rf24_pipe0_reading_address[5]; // last address set on pipe 0 for reading

// function prototypes
int rf24_init();
void rf24_set_channel(uint8_t channel);
uint8_t rf24_get_channel();
int  rf24_set_data_rate(rf24_datarate_e speed);
rf24_datarate_e rf24_get_data_rate();
void rf24_set_crc_length(rf24_crclength_e length);
rf24_crclength_e rf24_get_crc_length();
void rf24_disable_crc();
void rf24_set_amplifier_level(rf24_pa_dbm_e level);
rf24_pa_dbm_e rf24_get_amplifier_level();

void rf24_start_listening();
void rf24_stop_listening();
int  rf24_available(uint8_t* pipe_num); // checks if there are bytes available to read on a pipe
int rf24_read(uint8_t* buf, uint8_t len);
int  rf24_write(const uint8_t* buf, uint8_t len);
int rf24_start_write(const uint8_t* buf, uint8_t len);
void rf24_open_write_pipe(const uint8_t* address);
void rf24_open_read_pipe(uint8_t number, const uint8_t* address);   // read pipe number, and address

void rf24_power_up();
void rf24_power_down();
void rf24_toggle_features();
void rf24_set_retries(uint8_t delay, uint8_t count);


uint8_t rf24_read_register(uint8_t reg, uint8_t* buf, uint8_t len);
uint8_t rf24_read_single_register(uint8_t reg);
uint8_t rf24_write_register(uint8_t reg, const uint8_t* buf, uint8_t len);
uint8_t rf24_write_single_register(uint8_t reg, uint8_t value);
uint8_t rf24_write_payload(const uint8_t* buf, uint8_t data_len);
uint8_t rf24_read_payload(uint8_t* buf, uint8_t data_len);

uint8_t rf24_flush_tx();
uint8_t rf24_flush_rx();
uint8_t rf24_get_status();
void rf24_what_happened(uint8_t* tx_ok, uint8_t* tx_fail, uint8_t* rx_ready);

void rf24_enable_dynamic_payloads();
uint8_t rf24_get_dynamic_payload_size();
void rf24_enable_ack_payload();

int rf24_test_carrier();
int rf24_test_rpd();
void rf24_set_auto_ack_all(int enable);
void rf24_set_auto_ack(uint8_t pipe, int enable);
void rf24_write_ack_payload(uint8_t pipe, uint8_t* buf, uint8_t len);

// debug functions
void rf24_print_address_register(char* buf, const char* name, uint8_t reg, uint8_t qty);
void rf24_print_byte_register(char* dest, const char* name, uint8_t reg, uint8_t qty);
void rf24_print_status(char* dest, uint8_t status);
void rf24_print_details(char* dest);

uint8_t rf24_spi_transfer(uint8_t val);
void rf24_delay_ms(unsigned long ms);
void rf24_delay_us(unsigned long us);

#endif